"""
NRV-Cellular Level simulations
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq
(c) ETIS - University Cergy-Pontoise - CNRS
"""

from ...nmod.axons import *
from ...nmod.unmyelinated import *
from ...nmod.myelinated import *

from ...fmod.extracellular import *
from ...fmod.electrodes import *
from ...fmod.materials import *
from ...fmod.stimulus import *

from ...backend.MCore import *
from ...backend.log_interface import rise_error, rise_warning, pass_info

from ..saving_handler import *
from .CL_postprocessing import *
from .CL_discretization import *

unmyelinated_models=['HH','Rattay_Aberham','Sundt','Tigerholm','Schild_94','Schild_97']
myelinated_models=['MRG','Gaines_motor','Gaines_sensory']
thin_myelinated_models=['extended_Gaines','RGK']

def firing_threshold_point_source(diameter,L,dist_elec,cath_time = 100e-3,model='MRG',amp_max=2000,amp_tol=1,verbose=True,**kwargs):

	"""
	Find the firing threshold with point source approximation using binary search.

	Parameters
	----------
	diameter        : float
		axon diameter in um
	L               : float
		axon length in um
	dist_elec       : float
		y coordinate of the point source electrode in um
	cath_time      : float
		cathodic pulse width in ms
	model           : str
		Axon model
	amp_max         : float
		Maximum tested blocking amplitude for the binary search in uA
	amp_tol         : float
		Threshold tolerance in % for the binary search
	verbose         : bool
		set the verbosity of the search
	**kwargs:
		See below

	Keyword Arguments:
		material       :   str
			material used to compute the extracellular field
		position_elect : float 
			x location of the electrode, between 0 and 1
		z_elect        : float
			z coordinate of the electrode
		amp_min        : float
			minimum amplitude for the binary search
		f_dlambda      : float
			To set the number of segment with the d_dlambda rule (Hz)
		t_sim          : float
			Duration of the simulation in ms
		amp_tol_abs    : float
			Specify an absolute amplitude tolerance for the binary search
			in uA
		anod_first  : bool
			Set the anodic phase before the cathodic phase
		t_inter : float
			Specify the interphase delay in ms
		cath_an_ratio: float
			Specify the cathodic/anodic ratio
		dt              : float
			Set the dt to a specific value in ms
		cath_an_ratio          : float
			Set the tolorance for dt in %
		nseg            : int
			Set the number of segment for the axon
		nseg_tol        : float
			Set the nseg tolerance in %
		dt_tol        : float
			Set the dt tolerance in %
		amp_tol_abs     : float
			Set the absolute tolerance for the binary search in uA.
		node_shift      : float between -1 and 1
			Align electrode with Node of Ranvier. When Shift = 0, Electrode is aligned with node
		n_nodes         : integer
			Specify the number of Node of Ranvier for myelinated models. Overwrite L when specified.

	Returns
	-------
	threshold       : float
		estimated firing threshold in uA
	"""

	if 't_sim' in kwargs:
		t_sim = kwargs.get("t_sim")
	else :
		if (model in myelinated_models):
			t_sim = 5
		else:
			t_sim = 20

	if 'material' in kwargs:
		material = kwargs.get("material")
	else:
		material = material = 'endoneurium_bhadra'
	
	if 'anod_first' in kwargs:
		anod_first = kwargs.get("anod_first")
	else:
		anod_first = False

	if 't_inter' in kwargs:
		t_inter = kwargs.get("t_inter")
	else:
		t_inter = 0
	 
	if 'cath_an_ratio' in kwargs:
		cath_an_ratio = kwargs.get("cath_an_ratio")
	else:
		cath_an_ratio = 1
	
	if 'position_elec' in kwargs:
		position_elec = kwargs.get("position_elec")
	else:
		position_elec = 0.5

	if 'amp_min' in kwargs:
		amp_min = kwargs.get("amp_min")
	else:
		amp_min = 0

	if 'f_dlambda' in kwargs:
		f_dlambda = kwargs.get("f_dlambda")
	else:
		f_dlambda = 0

	if 'dt' in kwargs:
		dt = kwargs.get("dt")
	else:
		dt = 0

	if 'dt_tol' in kwargs:
		dt_tol = kwargs.get("dt_tol")
	else:
		dt_tol = amp_tol

	if 'nseg' in kwargs:
		nseg = kwargs.get("nseg")
	else:
		nseg = 0
    
	if 'Nseg_per_sec' in kwargs:
		Nseg_per_sec = kwargs.get("Nseg_per_sec")
	else:
		Nseg_per_sec = 0
		
	if 'nseg_tol' in kwargs:
		nseg_tol = kwargs.get("nseg_tol")
	else:
		nseg_tol = amp_tol

	if 'amp_tol_abs' in kwargs:
		amp_tol_abs = kwargs.get("amp_tol_abs")
	else :
		amp_tol_abs = 0

	if 'node_shift' in kwargs:
		node_shift = kwargs.get("node_shift")
	else :
		node_shift = 0
	
	if 'n_nodes' in kwargs:
		n_nodes = kwargs.get("n_nodes")
	else:
		n_nodes = 0
		
	amplitude_max_th=amp_max
	amplitude_min_th=amp_min
	amplitude_tol=amp_tol
	# axon
	y = 0
	z = 0

	if (dt == 0):
		if (cath_an_ratio>0 and cath_an_ratio<1):
			dt = Get_dt(dt_tol,model,simulation_category='Spike_threshold',stim_pw = cath_time/cath_an_ratio)
		else:
			dt = Get_dt(dt_tol,model,simulation_category='Spike_threshold',stim_pw = cath_time)

	if ((nseg == 0) and (f_dlambda == 0) and (Nseg_per_sec == 0)):
		if (model in unmyelinated_models):
			Nseg_per_sec = Get_nseg(nseg_tol,model,simulation_category='Spike_threshold',d = diameter,L=L)
		else:
			Nseg_per_sec = Get_nseg(nseg_tol,model,simulation_category='Spike_threshold')


	# extra cellular
	extra_material = load_material(material)
	#Dichotomy initialization
	previous_amp=amp_min
	delta_amp=np.abs(amp_max-amp_min)
	current_amp=amp_max
	Niter = 1
	keep_going = 1
	while (keep_going):
		if (verbose):
			pass_info('Iteration number '+str(Niter)+', testing firing current amplitude '+str(current_amp)+' uA')
		# create axon
		if model in unmyelinated_models:
			if Nseg_per_sec:
				axon1 = unmyelinated(y,z,diameter,L,dt=dt,model=model,Nseg_per_sec=Nseg_per_sec) #freq=f_dlambda
			else:
				axon1 = unmyelinated(y,z,diameter,L,dt=dt,freq=f_dlambda,model=model)
		elif model in myelinated_models:
			if (n_nodes>0):
				L=get_length_from_nodes(diameter,n_nodes)      
			if Nseg_per_sec:
				axon1 = myelinated(y,z,diameter,L,rec='nodes',dt=dt,model=model,Nseg_per_sec=Nseg_per_sec) #freq=f_dlambda
			else:
				axon1 = myelinated(y,z,diameter,L,rec='nodes',dt=dt,freq=f_dlambda,model=model)
		elif model in thin_myelinated_models:
			axon1 = thin_myelinated(y,z,diameter,L,rec='nodes',dt=dt,freq=f_dlambda,model=model)
		else:
			rise_error('Error: Specified model is not recognized.')
		# extra-cellular stimulation 
		x_elec = L*position_elec
		if (model in myelinated_models):         #Align electrode with node of Ranvier
			n_nodes = len(axon1.x_nodes)
			nearest_node = np.int32(np.round(position_elec*n_nodes))
			x_elec = axon1.x_nodes[nearest_node]
			if (node_shift!=0):                 #shift it if needed
				if (node_shift>0):
					if (nearest_node<n_nodes-1):
						x_next_node = axon1.x_nodes[nearest_node+1]
						d_internode = np.abs(x_next_node-x_elec)
						x_elec = x_elec+d_internode*node_shift
					else:
						rise_warning("Electrode is at the end of the axon and can not be further shifted.")
				else:
					if (nearest_node!=0):
						x_next_node = axon1.x_nodes[nearest_node+1]
						d_internode = np.abs(x_next_node-x_elec)
						x_elec = x_elec-d_internode*node_shift
					else:
						rise_warning("Electrode is at the end of the axon and can not be further shifted.")

		y_elec = dist_elec
		z_elec = 0
		elec_1 = point_source_electrode(x_elec,y_elec,z_elec)

		# stimulus def
		stim_1 = stimulus()
		start = 1
		I_cathod = current_amp
		if (cath_an_ratio>0):
			I_anod = I_cathod/cath_an_ratio

			stim_1.biphasic_pulse(start, I_cathod, cath_time, I_anod, t_inter,anod_first=anod_first)
		else:
			stim_1.biphasic_pulse(start, I_cathod, cath_time, 0, 0,anod_first=anod_first)

		stim_extra = stimulation(extra_material)
		stim_extra.add_electrode(elec_1,stim_1)
		axon1.attach_extracellular_stimulation(stim_extra)
		# simulate axon activity
		results = axon1.simulate(t_sim=t_sim)
		del axon1
		if (verbose):
			pass_info('... Iteration simulation performed in '+str(results['sim_time'])+' s')
		# post-process results
		rasterize(results,'V_mem')
		delta_amp=np.abs(current_amp-previous_amp)
		if (amp_tol_abs>0):
			if (delta_amp<amp_tol_abs):
				keep_going = 0
			else:
				keep_going = 1
		else:
			if (current_amp>previous_amp):
				tol = 100*delta_amp/current_amp
			else:
				tol = 100*delta_amp/previous_amp
			if (tol <= amp_tol):
				keep_going = 0
			else:
				keep_going = 1
		previous_amp=current_amp
		# test simulation results, update dichotomy
		if (len(results['V_mem_raster_position'])>0):
			if (current_amp==amp_min):
				rise_warning("Minimum Stimulation Current is too High!")
				return (0)
			if (verbose):
				pass_info("... Spike triggered")
			amplitude_max_th=previous_amp
			current_amp=(delta_amp/2)+amplitude_min_th
		else:
			if (current_amp==amp_max):
				rise_warning("Maximum Stimulation Current is too Low!")
				return (0)
			if (verbose):
				pass_info("... Spike not triggered")
			current_amp=amplitude_max_th-delta_amp/2
			amplitude_min_th=previous_amp

		if (previous_amp==amp_max):
			current_amp=amp_min

		Niter += 1
	return current_amp

def firing_threshold_from_axon(axon,cath_time= 100e-3,amp_max=2000,amp_tol=1,verbose=True,**kwargs):
	
	"""
	Find the firing threshold from a specified axon using binary search

	Parameters
	----------
	axon        : axon class
		axon class 
	cath_time      : float
		cathodic pulse width in ms
	amp_max         : float
		Maximum tested blocking amplitude for the binary search in uA
	amp_tol         : float
		Threshold tolerance in % for the binary search
	verbose         : bool
		set the verbosity of the search
	**kwargs:
		See below

	Keyword Arguments:
		amp_min        : float
			minimum amplitude for the binary search
		t_sim          : float
			Duration of the simulation in ms
		amp_tol_abs    : float
			Specify an absolute amplitude tolerance for the binary search
			in uA
		dt             : float
			Set the dt to a specific value in ms
		anod_first  : bool
			Set the anodic phase before the cathodic phase
		t_inter : float
			Specify the interphase delay in ms
		cath_an_ratio: float
			Specify the cathodic/anodic ratio
		cath_an_ratio          : float
			Set the tolorance for dt in %
		amp_tol_abs     : float
			Set the absolute tolerance for the binary search in uA.
		dt_tol        : float
			Set the dt tolerance in %

	Returns
	-------
	threshold       : float
		estimated firing threshold in uA
	"""


	if 't_sim' in kwargs:
		t_sim = kwargs.get("t_sim")
	else :
		t_sim = 10
	
	if 'anod_first' in kwargs:
		anod_first = kwargs.get("anod_first")
	else:
		anod_first = False

	if 't_inter' in kwargs:
		t_inter = kwargs.get("t_inter")
	else:
		t_inter = 0
	 
	if 'cath_an_ratio' in kwargs:
		cath_an_ratio = kwargs.get("cath_an_ratio")
	else:
		cath_an_ratio = 1
	
	if 'amp_min' in kwargs:
		amp_min = kwargs.get("amp_min")
	else:
		amp_min = 0

	if 'amp_tol_abs' in kwargs:
		amp_tol_abs = kwargs.get("amp_tol_abs")
	else :
		amp_tol_abs = 0
	
	if 'dt' in kwargs:
		dt = kwargs.get("dt")
	else:
		dt = 0

	if 'dt_tol' in kwargs:
		dt_tol = kwargs.get("dt_tol")
	else:
		dt_tol = amp_tol

	if (dt == 0):
		if (cath_an_ratio>0 and cath_an_ratio<1):
			dt = Get_dt(dt_tol,axon.model,simulation_category='Spike_threshold',stim_pw = cath_time/cath_an_ratio)
		else:
			dt = Get_dt(dt_tol,axon.model,simulation_category='Spike_threshold',stim_pw = cath_time)
		
	amplitude_max_th=amp_max
	amplitude_min_th=amp_min
	amplitude_tol=amp_tol

	#Dichotomy initialization
	previous_amp=amp_min
	delta_amp=np.abs(amp_max-amp_min)
	current_amp=amp_max
	Niter = 1
	keep_going = 1
	
	#axon.save_axon(save=True, fname='_axon.json', extracel_context=True)
	#del (axon)
	while (keep_going):
		if (verbose):
			pass_info('Iteration number '+str(Niter)+', testing firing current amplitude '+str(current_amp)+' uA')

		# stimulus def
		stim_1 = stimulus()
		start = 1
		I_cathod = current_amp
		#I_cathod = 
		if (cath_an_ratio>0):
			I_anod = I_cathod/cath_an_ratio
			stim_1.biphasic_pulse(start, I_cathod, cath_time, I_anod, t_inter,anod_first=anod_first)
		else:
			stim_1.biphasic_pulse(start, I_cathod, cath_time, 0, 0,anod_first=anod_first)
		#axon_load = load_any_axon(axon, extracel_context=True)
		#axon_th = copy.deepcopy(axon)
		axon.dt = dt
		axon.change_stimulus_from_elecrode(0, stim_1)

		# simulate axon activity
		results = axon.simulate(t_sim=t_sim,loaded_footprints=True)
		#del (axon)
		if (verbose):
			pass_info('... Iteration simulation performed in '+str(results['sim_time'])+' s')
		# post-process results
		rasterize(results,'V_mem')
		delta_amp=np.abs(current_amp-previous_amp)
		if (amp_tol_abs>0):
			if (delta_amp<amp_tol_abs):
				keep_going = 0
			else:
				keep_going = 1
		else:
			if (current_amp>previous_amp):
				tol = 100*delta_amp/current_amp
			else:
				tol = 100*delta_amp/previous_amp
			if (tol <= amp_tol):
				keep_going = 0
			else:
				keep_going = 1
		previous_amp=current_amp
		# test simulation results, update dichotomy
		if (len(results['V_mem_raster_position'])>0):
			if (current_amp==amp_min):
				rise_warning("Minimum Stimulation Current is too High!")
				return (0)
			if (verbose):
				pass_info("... Spike triggered")
			amplitude_max_th=previous_amp
			current_amp=(delta_amp/2)+amplitude_min_th
		else:
			if (current_amp==amp_max):
				rise_warning("Maximum Stimulation Current is too Low!")
				return (0)
			if (verbose):
				pass_info("... Spike not triggered")
			current_amp=amplitude_max_th-delta_amp/2
			amplitude_min_th=previous_amp

		if (previous_amp==amp_max):
			current_amp=amp_min

		Niter += 1
	del (axon)
	return current_amp

def para_firing_threshold(diameter,L,material,dist_elec,cath_first=True,cath_time = 60e-3,t_inter = 40e-3,cath_an_ratio = 5,position_elec=0.5,model='MRG',amp_max=2000,amp_min=0,amp_tol=1,verbose=False,f_dlambda = 100,dt=0.005):
	if MCH.is_alone() or MCH.size < 3:
		if MCH.do_master_only_work():
			rise_error('Error: parallel evaluation of a threshold by binary search needs at least 3 parallel processes, but only', MCH.size, 'launched')
		sys.exit(1)
	amplitude_max_th=amp_max
	amplitude_min_th=amp_min
	amplitude_tol=amp_tol
	# axon
	y = 0
	z = 0
	# extra cellular
	extra_material = load_material(material)
	#Dichotomy initialization
	Niter = 1
	values = np.linspace(amplitude_min_th, amplitude_max_th, num=MCH.size)
	delta_amp = values[-1] - values[0]
	current_amp = (values[-1] - values[0])/2
	all_values = values
	while (delta_amp>amplitude_tol or Niter == 1):
		# update toledance
		delta_amp = values[1] - values[0]
		# split job
		current_amp = values[MCH.rank]
		# create axon
		if model in unmyelinated_models:
			axon1 = unmyelinated(y,z,diameter,L,dt=dt,freq=f_dlambda,model=model)
		elif model in myelinated_models:
			axon1 = myelinated(y,z,diameter,L,rec='nodes',dt=dt,freq=f_dlambda,model=model)
		elif model in thin_myelinated_models:
			axon1 = thin_myelinated(y,z,diameter,L,rec='nodes',dt=dt,freq=f_dlambda,model=model)
		else:
			axon1 = myelinated(y,z,diameter,L,rec='nodes',dt=dt,freq=f_dlambda,model=model)
		# extra-cellular stimulation
		x_elec = L*position_elec
		y_elec = dist_elec
		z_elec = 0
		elec_1 = point_source_electrode(x_elec,y_elec,z_elec)

		# stimulus def
		stim_1 = stimulus()
		start = 1
		I_cathod = current_amp
		I_anod = I_cathod/cath_an_ratio
		stim_1.biphasic_pulse(start, I_cathod, cath_time, I_anod, t_inter,anod_first=(not cath_first))

		stim_extra = stimulation(extra_material)
		stim_extra.add_electrode(elec_1,stim_1)
		axon1.attach_extracellular_stimulation(stim_extra)
		# simulate axon activity
		results = axon1.simulate(t_sim=5)
		del axon1
		# post-process results
		rasterize(results,'V_mem')
		# test simulation results, gather results to master
		if (len(results['V_mem_raster_position'])>0):
			spike = np.asarray([True])
		else:
			spike = np.asarray([False])
		all_spikes = MCH.gather_jobs_as_array(spike)
		if MCH.do_master_only_work():
			# add the extrema (False at start, True at the end) already computed at the previous step if Niter > 1
			if Niter > 1:
				all_spikes = np.insert(np.append(all_spikes, True), 0, False)
			# find the zone where the spike appears, update min max
			appear_index = np.argmax(all_spikes==True)
			amplitude_max_th = all_values[appear_index]
			amplitude_min_th = all_values[appear_index-1]
			# update current amp
			current_amp = amplitude_min_th + (amplitude_max_th-amplitude_min_th)/2
			# compute new values
			all_values = np.linspace(amplitude_min_th, amplitude_max_th, num=MCH.size+3)
			values = all_values[1:-1]
		# share new values
		values = MCH.master_broadcasts_array_to_all(values)
		Niter += 1
	# adapt Niter for correct return
	Niter -= 1
	# share master current amp to all process
	current_amp = MCH.master_broadcasts_array_to_all(current_amp)
	return current_amp, Niter

def blocking_threshold_point_source(diameter,L,dist_elec,block_freq,model='MRG',amp_max=2000,amp_tol=1,dt=0.005,Nseg_per_sec=2,verbose=True,**kwargs):
	"""
	Find the blocking threshold with point source approximation using binary search.

	Parameters
	----------
	diameter        : float
		axon diameter in um
	L               : float
		axon length in um
	dist_elec       : float
		y coordinate of the point source electrode in um
	block_freq      : float
		Blocking frequency in kHz
	model           : str
		Axon model
	amp_max         : float
		Maximum tested blocking amplitude for the binary search in uA
	amp_tol         : float
		Threshold tolerance in % for the binary search
	dt              : float
		time discretization in ms
	Nseg_per_sec    : int
		Number of segment per section (myelinated axons)
		Number of segment per mm of length (unmyelinated axons)
	verbose         : bool
		set the verbosity of the search

	**kwargs:
		See below

	Keyword Arguments:
		material       :   str
			material used to compute the extracellular field
		position_elect : float 
			x location of the electrode, between 0 and 1
		z_elect        : float
			z coordinate of the electrode
		amp_min        : float
			minimum amplitude for the binary search
		f_dlambda      : float
			To set the number of segment with the d_dlambda rule (Hz)
		t_sim          : float
			Duration of the simulation in ms
		amp_tol_abs    : float
			Specify an absolute amplitude tolerance for the binary search
			in uA
		t_position     : float
			Positition on the test electrode. 
			Relative position for unmyelinated axons.
			Node number for myelinated axons.
		t_start        : float
			start of the test pulse, in ms
		t_duration     : float
			duration of the test pulse, in ms 
		t_amplitude    : float
			Amplitude of the test pulse, in nA
		b_start        : float
			start of the blocking stimulation in ms
		b_duration     : float 
			duration of the blocking stimulation in ms
		node_shift      : float between -1 and 1
			Align electrode with Node of Ranvier. When Shift = 0, Electrode is aligned with node
		n_nodes         : integer
			Specify the number of Node of Ranvier for myelinated models. Overwrite L when specified.

	Returns
	-------
	threshold       : float
		estimated threshold in uA
	"""

	if 'material' in kwargs:
		material = kwargs.get("material")
	else :
		material = 'endoneurium_bhadra'

	if 'position_elec' in kwargs:
		position_elec = kwargs.get("position_elec")
	else :
		position_elec = 0.5

	if 'z_elec' in kwargs:
		z_elec = kwargs.get("z_elec")
	else :
		z_elec = 0

	if 'amp_min' in kwargs:
		amp_min = kwargs.get("amp_min")
	else :
		amp_min = 0

	if 'f_dlambda' in kwargs:
		f_dlambda = kwargs.get("f_dlambda")
	else :
		f_dlambda = 100

	if 't_sim' in kwargs:
		t_sim = kwargs.get("t_sim")
	else :
		t_sim = 40

	if 'amp_tol_abs' in kwargs:
		amp_tol_abs = kwargs.get("amp_tol_abs")
	else :
		amp_tol_abs = 0

	if  't_position' in kwargs:
		t_position = kwargs.get("t_position")
	else: 
		t_position = 0.05

	if 't_start' in kwargs:
		t_start = kwargs.get("t_start")
	else :
		t_start = 30
		
	if 't_duration' in kwargs:
		t_duration = kwargs.get("t_duration")
	else :
		t_duration = 1

	if 't_amplitude' in kwargs:
		t_amplitude = kwargs.get("t_amplitude")
	else :
		t_amplitude = 2

	if 'b_start' in kwargs:
		b_start = kwargs.get("b_start")
	else :
		b_start = 3

	if 'b_duration' in kwargs:
		b_duration = kwargs.get("b_duration")
	else :
		b_duration = t_sim
	
	if 'node_shift' in kwargs:
		node_shift = kwargs.get("node_shift")
	else :
		node_shift = 0

	if 'n_nodes' in kwargs:
		n_nodes = kwargs.get("n_nodes")
	else:
		n_nodes = 0

	amplitude_max_th=amp_max
	amplitude_min_th=amp_min
	amplitude_tol=amp_tol

	y = 0
	z = 0

	extra_material = load_material(material)
	#Dichotomy initialization
	previous_amp=amp_min
	delta_amp=np.abs(amp_max-amp_min)
	current_amp=amp_max
	Niter = 1
	keep_going = 1
	while (keep_going):
		if (verbose):
			pass_info('Iteration number '+str(Niter)+', testing block current amplitude '+str(current_amp)+' uA')
		# create axon
		if model in unmyelinated_models:
			if Nseg_per_sec:
				axon1 = unmyelinated(y,z,diameter,L,dt=dt,Nseg_per_sec=Nseg_per_sec,Nsec=1,model=model)
			else:
				axon1 = unmyelinated(y,z,diameter,L,dt=dt,freq=f_dlambda,model=model)
		elif model in myelinated_models:
			if (n_nodes>0):
				L=get_length_from_nodes(diameter,n_nodes)  
			if Nseg_per_sec: 
				axon1 = myelinated(y,z,diameter,L,rec='nodes',dt=dt,Nseg_per_sec=Nseg_per_sec,model=model)
			else:
				axon1 = myelinated(y,z,diameter,L,rec='nodes',dt=dt,freq=f_dlambda,model=model)
		elif model in thin_myelinated_models:
			axon1 = thin_myelinated(y,z,diameter,L,rec='nodes',dt=dt,freq=f_dlambda,model=model)
		else:
			if (n_nodes>0):
				L=get_length_from_nodes(diameter,n_nodes)  
			axon1 = myelinated(y,z,diameter,L,rec='nodes',dt=dt,freq=f_dlambda,model=model)
		# insert test spike
		axon1.insert_I_Clamp(t_position, t_start, t_duration, t_amplitude)
		# extra-cellular stimulation
		x_elec = L*position_elec
		if (model in myelinated_models):         #Align electrode with node of Ranvier
			n_nodes = len(axon1.x_nodes)
			nearest_node = np.int32(np.round(position_elec*n_nodes))
			x_elec = axon1.x_nodes[nearest_node]
			if (node_shift!=0):                 #shift it if needed
				if (node_shift>0):
					if (nearest_node<n_nodes-1):
						x_next_node = axon1.x_nodes[nearest_node+1]
						d_internode = np.abs(x_next_node-x_elec)
						x_elec = x_elec+d_internode*node_shift
					else:
						rise_warning("Electrode is at the end of the axon and can not be further shifted.")
				else:
					if (nearest_node!=0):
						x_next_node = axon1.x_nodes[nearest_node+1]
						d_internode = np.abs(x_next_node-x_elec)
						x_elec = x_elec-d_internode*node_shift
					else:
						rise_warning("Electrode is at the end of the axon and can not be further shifted.")
		y_elec = dist_elec
		z_elec = z_elec
		elec_1 = point_source_electrode(x_elec,y_elec,z_elec)
		stim_1 = stimulus()
		stim_1.sinus(b_start, b_duration, current_amp, block_freq ,dt=1/(block_freq*20))
		stim_extra = stimulation(extra_material)
		stim_extra.add_electrode(elec_1,stim_1)
		axon1.attach_extracellular_stimulation(stim_extra)
		# simulate axon activity
		results = axon1.simulate(t_sim=t_sim)
		del axon1
		if (verbose):
			pass_info('... Iteration simulation performed in '+str(results['sim_time'])+' s')
		# post-process results
		filter_freq(results,'V_mem',block_freq)
		rasterize(results,'V_mem_filtered',threshold = 0)
		delta_amp=np.abs(current_amp-previous_amp)
		if (amp_tol_abs>0):
			if (delta_amp<amp_tol_abs):
				keep_going = 0
			else:
				keep_going = 1
		else:
			if (current_amp>previous_amp):
				tol = 100*delta_amp/current_amp
			else:
				tol = 100*delta_amp/previous_amp
			if (tol <= amp_tol):
				keep_going = 0
			else:
				keep_going = 1
		previous_amp=current_amp
		# test simulation results, update dichotomy
		if block(results,t_start=t_start) == False:
			if (current_amp == amp_max):
				rise_warning("Maximum Stimulation Current is too Low!")
				return (0)
			if (verbose):
				pass_info("... Spike not blocked")
			amplitude_min_th = current_amp
			current_amp = (delta_amp/2)+amplitude_min_th
		else:
			if (current_amp == amp_min):
				rise_warning("Minimum Stimulation Current is too High!")
				return (0)
			if (verbose):
				pass_info("... Spike blocked")
			amplitude_max_th = current_amp
			current_amp = amplitude_max_th-delta_amp/2

		if (previous_amp == amp_max):
				current_amp = amp_min
		Niter += 1
	return current_amp

def blocking_threshold_from_axon(axon,block_freq=10,amp_max=2000,amp_tol=1,dt=0.005,verbose=True,**kwargs):
	"""
	Find the blocking threshold with point source approximation using binary search.

	Parameters
	----------
	diameter        : float
		axon diameter in um
	L               : float
		axon length in um
	dist_elec       : float
		y coordinate of the point source electrode in um
	block_freq      : float
		Blocking frequency in kHz
	model           : str
		Axon model
	amp_max         : float
		Maximum tested blocking amplitude for the binary search in uA
	amp_tol         : float
		Threshold tolerance in % for the binary search
	dt              : float
		time discretization in ms
	Nseg_per_sec    : int
		Number of segment per section (myelinated axons)
		Number of segment per mm of length (unmyelinated axons)
	verbose         : bool
		set the verbosity of the search

	**kwargs:
		See below

	Keyword Arguments:
		material       :   str
			material used to compute the extracellular field
		position_elect : float 
			x location of the electrode, between 0 and 1
		z_elect        : float
			z coordinate of the electrode
		amp_min        : float
			minimum amplitude for the binary search
		f_dlambda      : float
			To set the number of segment with the d_dlambda rule (Hz)
		t_sim          : float
			Duration of the simulation in ms
		amp_tol_abs    : float
			Specify an absolute amplitude tolerance for the binary search
			in uA
		t_position     : float
			Positition on the test electrode. 
			Relative position for unmyelinated axons.
			Node number for myelinated axons.
		t_start        : float
			start of the test pulse, in ms
		t_duration     : float
			duration of the test pulse, in ms 
		t_amplitude    : float
			Amplitude of the test pulse, in nA
		b_start        : float
			start of the blocking stimulation in ms
		b_duration     : float 
			duration of the blocking stimulation in ms
		node_shift      : float between -1 and 1
			Align electrode with Node of Ranvier. When Shift = 0, Electrode is aligned with node
		n_nodes         : integer
			Specify the number of Node of Ranvier for myelinated models. Overwrite L when specified.

	Returns
	-------
	threshold       : float
		estimated threshold in uA
	"""

	if 'amp_min' in kwargs:
		amp_min = kwargs.get("amp_min")
	else :
		amp_min = 0

	if 'f_dlambda' in kwargs:
		f_dlambda = kwargs.get("f_dlambda")
	else :
		f_dlambda = 100

	if 't_sim' in kwargs:
		t_sim = kwargs.get("t_sim")
	else :
		t_sim = 40

	if 'amp_tol_abs' in kwargs:
		amp_tol_abs = kwargs.get("amp_tol_abs")
	else :
		amp_tol_abs = 0

	if  't_position' in kwargs:
		t_position = kwargs.get("t_position")
	else: 
		t_position = 0.05

	if 't_start' in kwargs:
		t_start = kwargs.get("t_start")
	else :
		t_start = 30
		
	if 't_duration' in kwargs:
		t_duration = kwargs.get("t_duration")
	else :
		t_duration = 1

	if 't_amplitude' in kwargs:
		t_amplitude = kwargs.get("t_amplitude")
	else :
		t_amplitude = 2

	if 'b_start' in kwargs:
		b_start = kwargs.get("b_start")
	else :
		b_start = 3

	if 'b_duration' in kwargs:
		b_duration = kwargs.get("b_duration")
	else :
		b_duration = t_sim
	
	amplitude_max_th=amp_max
	amplitude_min_th=amp_min
	amplitude_tol=amp_tol

	#Dichotomy initialization
	previous_amp=amp_min
	delta_amp=np.abs(amp_max-amp_min)
	current_amp=amp_max
	Niter = 1
	keep_going = 1
	while (keep_going):
		if (verbose):
			pass_info('Iteration number '+str(Niter)+', testing block current amplitude '+str(current_amp)+' uA')
		
		# insert test spike
		axon.insert_I_Clamp(t_position, t_start, t_duration, t_amplitude)
		# extra-cellular stimulation

		stim_1 = stimulus()
		stim_1.sinus(b_start, b_duration, current_amp, block_freq ,dt=1/(block_freq*20))

		axon.dt = dt
		axon.change_stimulus_from_elecrode(0, stim_1)

		# simulate axon activity
		results = axon.simulate(t_sim=t_sim,loaded_footprints=True)

		if (verbose):
			pass_info('... Iteration simulation performed in '+str(results['sim_time'])+' s')
		# post-process results
		filter_freq(results,'V_mem',block_freq)
		rasterize(results,'V_mem_filtered',threshold = 0)
		delta_amp=np.abs(current_amp-previous_amp)
		if (amp_tol_abs>0):
			if (delta_amp<amp_tol_abs):
				keep_going = 0
			else:
				keep_going = 1
		else:
			if (current_amp>previous_amp):
				tol = 100*delta_amp/current_amp
			else:
				tol = 100*delta_amp/previous_amp
			if (tol <= amp_tol):
				keep_going = 0
			else:
				keep_going = 1
		previous_amp=current_amp
		# test simulation results, update dichotomy
		if block(results,t_start=t_start) == False:
			if (current_amp == amp_max):
				rise_warning("Maximum Stimulation Current is too Low!")
				return (0)
			if (verbose):
				pass_info("... Spike not blocked")
			amplitude_min_th = current_amp
			current_amp = (delta_amp/2)+amplitude_min_th
		else:
			if (current_amp == amp_min):
				rise_warning("Minimum Stimulation Current is too High!")
				return (0)
			if (verbose):
				pass_info("... Spike blocked")
			amplitude_max_th = current_amp
			current_amp = amplitude_max_th-delta_amp/2

		if (previous_amp == amp_max):
				current_amp = amp_min
		Niter += 1
	return current_amp

def para_blocking_threshold(diameter,L,material,dist_elec,block_freq,position_elec=0.5,model='MRG',amp_max=2000,amp_min=0,amp_tol=15,verbose=True,f_dlambda = 100,dt=0.005):
	amplitude_max_th=amp_max
	amplitude_min_th=amp_min
	amplitude_tol=amp_tol
	# axon
	y = 0
	z = 0
	# test spike
	t_start = 20
	duration = 0.1
	amplitude = 3
	# extra cellular
	extra_material = load_material(material)
	block_start=3 #ms
	block_duration=20 #ms
	#Dichotomy initialization
	Niter = 1
	values = np.linspace(amplitude_min_th, amplitude_max_th, num=MCH.size)
	delta_amp = values[-1] - values[0]
	current_amp = (values[-1] - values[0])/2
	all_values = values
	while (delta_amp>amplitude_tol or Niter == 1):
		# update toledance
		delta_amp = values[1] - values[0]
		# split job
		current_amp = values[MCH.rank]
		# create axon
		if model in unmyelinated_models:
			axon1 = unmyelinated(y,z,diameter,L,dt=dt,freq=f_dlambda,model=model)
		elif model in myelinated_models:
			axon1 = myelinated(y,z,diameter,L,rec='nodes',dt=dt,freq=f_dlambda,model=model)
		elif model in thin_myelinated_models:
			axon1 = thin_myelinated(y,z,diameter,L,rec='nodes',dt=dt,freq=f_dlambda,model=model)
		else:
			axon1 = myelinated(y,z,diameter,L,rec='nodes',dt=dt,freq=f_dlambda,model=model)
		# insert test spike
		axon1.insert_I_Clamp(0, t_start, duration, amplitude)
		# extra-cellular stimulation
		x_elec = L*position_elec
		y_elec = dist_elec
		z_elec = 0
		elec_1 = point_source_electrode(x_elec,y_elec,z_elec)
		stim_1 = stimulus()
		stim_1.sinus(block_start, block_duration, current_amp, block_freq ,dt=1/(block_freq*20))
		stim_extra = stimulation(extra_material)
		stim_extra.add_electrode(elec_1,stim_1)
		axon1.attach_extracellular_stimulation(stim_extra)
		# simulate axon activity
		results = axon1.simulate(t_sim=25)
		del axon1
		# post-process results
		rasterize(results,'V_mem')

		# test simulation results, gather results to master
		blocked = [block(results)]
		all_blocks = MCH.gather_jobs_as_array(blocked)
		if MCH.do_master_only_work():
			# add the extrema (False at start, True at the end) already computed at the previous step if Niter > 1
			if Niter > 1:
				all_blocks = np.insert(np.append(all_blocks, True), 0, False)
			# find the zone where the blocking appears, update min max
			appear_index = np.argmax(all_blocks==True)
			amplitude_max_th = all_values[appear_index]
			amplitude_min_th = all_values[appear_index-1]
			# update current amp
			current_amp = amplitude_min_th + (amplitude_max_th-amplitude_min_th)/2
			# compute new values
			all_values = np.linspace(amplitude_min_th, amplitude_max_th, num=MCH.size+3)
			values = all_values[1:-1]
		# share new values
		values = MCH.master_broadcasts_array_to_all(values)
		Niter += 1
	# adapt Niter for correct return
	Niter -= 1
	# share master current amp to all process
	current_amp = MCH.master_broadcasts_array_to_all(current_amp)
	return current_amp, Niter



################################
## rasterize membrane voltage ##
################################
nrv.rasterize(sim_results,'V_mem')

###############################
## remove non nevessary data ##
###############################
list_keys = ['ID', 'L', 'V_mem_raster_position', 'V_mem_raster_x_position', 'V_mem_raster_time_index', \
'V_mem_raster_time', 'myelinated', 'y', 'z','diameter']
removable_keys = []
for key in sim_results:
    if not key in list_keys:
        removable_keys += [key]
for key in removable_keys:
    nrv.remove_key(sim_results,key, verbose=False)

####################################
## check if the axon fired or not ##
####################################
file_object = open(folder_name+'/spike_summary.csv', 'a')
if len(sim_results['V_mem_raster_position']) == 0:
	# no spike
	spike = 0
else:
	spike = 1
line = str(sim_results['ID'])+'\t'+\
		str(sim_results['y'])+'\t'+\
		str(sim_results['z'])+'\t'+\
		str(sim_results['diameter'])+'\t'+\
		str(sim_results['myelinated'])+'\t'+\
		str(spike)+'\n'
file_object.write(line)
file_object.close()


import json
import numpy as np

from ....backend.file_handler import json_load, json_dump, rmv_ext
from ....backend.log_interface import rise_error, rise_warning, pass_info
from ....backend.NRV_Class import NRV_class
def is_sim_param(X):
    """
    check if an object is a SimParameters, return True if yes, else False

    Parameters
    ----------
    X : object
        object to test

    Returns
    -------
    bool
        True it the type is a SimParameters object
    """
    return isinstance(X, SimParameters)


class SimParameters(NRV_class):
    """
    Class gathering parameters of a FEM FEniCS simulation. It allows to add,
    modify and store domains, boundaries condition and internal boundaries (thin layers)
    Can be saved and used to generate FEMsimulation.
    """
    def __init__(self, D=3, mesh_file="", data=None):
        """
        initialisation of the SimParameters:
        
        Parameters
        ----------
        D               :int
            dim of the mesh, by default 3
            NB: only 3 is implemented
        mesh_file       :str
            mesh directory and file name: by default ""
        data            :str, dict or SimParameters
            if not None, load SimParameters attribute from data, by default None
            (see SimParameters.load)
        """
        super().__init__()
        self.type = "simparameters"
        self.D = D
        self.mesh_file = rmv_ext(mesh_file)

        # domains
        self.Ndomains = 0
        self.domainsID = []
        self.domains_list = {}

        # boundaries
        self.Nboundaries = 0
        self.boundariesID = []
        self.boundaries_list = {}

        # internal boundaries (thin layer)
        self.inbound = False
        self.Ninboundaries = 0
        self.inboundariesID = []
        self.inboundaries_list = {}

        # gather mat_pty name for domain and internal layer
        self.mat_pty_map = {}

        if data is not None:
            self.load(data)


    def save(self, save=False, fname='SimParameters.json'):
        """
        Return SimParameters as dictionary and eventually save it as json file

        Parameters
        ----------
        save    : bool
            if True, save in json files
        fname   : str
            Path and Name of the saving file, by default 'SimParameters.json'

        Returns
        -------
        sp_dic : dict
            dictionary containing all information
        """
        sp_dic = {}
        sp_dic['D'] = self.D
        sp_dic['file'] = self.mesh_file
        sp_dic['Ndomains'] = self.Ndomains
        sp_dic['domainsID'] = self.domainsID
        sp_dic['domains'] = self.domains_list
        sp_dic['Nboundaries'] = self.Nboundaries
        sp_dic['boundariesID'] = self.boundariesID
        sp_dic['boundaries'] = self.boundaries_list
        sp_dic['inbound'] = self.inbound
        sp_dic['Ninboundaries'] = self.Ninboundaries
        sp_dic['inboundariesID'] = self.inboundariesID
        sp_dic['inboundaries'] = self.inboundaries_list
        sp_dic['mat_pty_map'] = self.mat_pty_map
        if save:
            json_dump(sp_dic, fname)
        return sp_dic


    def load(self, data):
        """
        Load all SimParameters properties from a dictionary or a json file

        Parameters
        ----------
        data    : str or dict
            json file path or dictionary containing SimParameters information
        """
        if type(data) == str:
            sp_dic = json_load(data)
        elif is_sim_param(data):
            sp_dic = self.save()
        else: 
            sp_dic = data
        self.D = sp_dic['D']
        self.mesh_file = sp_dic['file']
        self.Ndomains = sp_dic['Ndomains']
        self.domainsID = sp_dic['domainsID']
        self.domains_list = sp_dic['domains']
        self.Nboundaries = sp_dic['Nboundaries']
        self.boundariesID = sp_dic['boundariesID']
        self.boundaries_list = sp_dic['boundaries']
        self.inbound = sp_dic['inbound']
        self.Ninboundaries = sp_dic['Ninboundaries']
        self.inboundariesID = sp_dic['inboundariesID']
        self.inboundaries_list = sp_dic['inboundaries']
        self.mat_pty_map = sp_dic['mat_pty_map']

    def save_SimParameters(self, save=False, fname='SimParameters.json'):
        rise_warning('save_SimParameters is a deprecated method use save')
        return self.save(save=save, fname=fname)
    def load_SimParameters(self, data='SimParameters.json'):
        rise_warning('load_SimParameters is a deprecated method use load')
        self.load(data=data)

    def set_mesh_file(self, new_mesh_file):
        """
        set a new mesh file name
        
        Parameters
        ----------
        new_mesh_file   :str 
            name of the new filename, will be save without the filename extension
        """
        self.mesh_file = new_mesh_file.replace('.gmsh','').replace('.xdml','') 

    def __update_ID_list(self, lname, ID):
        """
        Internal use only:
        Parameter
        """
        if lname =='domains':
            IDlist = self.domainsID
        elif lname =='inbound':
            IDlist = self.inboundariesID
        elif lname =='bound':
            IDlist = self.boundariesID
        else:
            rise_error('_update_ID_list failed due to unknow list type')
        
        if IDlist == []:
            if ID is None:
                IDlist = [1]
            else:
                IDlist = [ID]
        else:
            np.sort(IDlist)
            if ID is None:
                IDlist += [IDlist[-1]+1]
            else:
                if ID in IDlist:
                    return ID
                else:
                    IDlist += [ID]
        
        if lname =='domains':
            self.domainsID = IDlist
        elif lname =='inbound':
            self.inboundariesID = IDlist     
        else:
            self.boundariesID = IDlist
        return IDlist[-1]

    def add_domain(self, mesh_domain, mat_pty=None, mat_file=None, mat_perm=None, ID=None):
        """
        add new domain or update if mesh_domain already exists
        
        Parameters
        ----------
        mesh_domain     : int
            Mesh ID of the domain (should be 3D)
        mat_pty         : str, float or list[3]
            Material property, either mat_file or mat_perm
        mat_file        : str
            Material filename (see fmod.material.py) if None use mat_perm,
                by default None
        mat_perm        : float or list[3]
            Material permitivity, if float isotropic mat, if list[3] anisotropic mat
                if None set to 1 (and mat_file is None), by default None
        """
        if mat_pty is not None:
            mat_pty = mat_pty
        elif mat_file is not None:
            mat_pty = mat_file
        elif mat_perm is not None:
            mat_pty = mat_perm
        else:
            mat_pty = 1
            rise_warning('neither mat_file or mat_perm set, set to default permitivity 1 S/m')

        if not mesh_domain in self.domains_list:
            IDdom = self.__update_ID_list('domains', mesh_domain)
            self.domains_list[IDdom] = {'mesh_domain': mesh_domain, 'mat_pty':mat_pty, 'mixed_domain':[mesh_domain]}
        else:
            self.domains_list[mesh_domain]['mat_pty'] = mat_pty
        self.mat_pty_map[mesh_domain] = mat_pty
        self.Ndomains = len(self.domainsID)
    
    def add_boundary(self, mesh_domain, btype, value=None, variable=None, mesh_domain_3D=0, ID=None):
        """
        add new boundary or update if mesh_domain already exists
        
        Parameters
        ----------
        mesh_domain     : int
            Mesh ID of the boundary (should be 2D)
        btype           : str {'Dirichlet', 'Neuman'}
            Type of boundary condition either 'Dirichlet' to set the 
        value           : float
            set the condition to the value
        variable        : str
            name of a python variable from which the boundary value should be taken. 
            NB: variable should be place as kwarg in prepare_sim see examples
        mesh_domain_3D  : int
            ID of the mesh of corresponding volume, by default 0
            NB: Most of the time 0 except for LIFE electrode, must be set to the corresponding fascicle)
        ID              : int
            ID of the boundary condition
        """
        IDbound= self.__update_ID_list('bound', mesh_domain)
        if value is not None:
            self.boundaries_list[IDbound] = {'mesh_domain': mesh_domain, 'condition':btype,\
                "value":value, "mesh_domain_3D":mesh_domain_3D}
        elif variable is not None:
            self.boundaries_list[IDbound] = {'mesh_domain': mesh_domain, 'condition':btype,\
                "variable":variable, "mesh_domain_3D":mesh_domain_3D}
        else: 
            rise_warning("boundary not set, variable or boundary have to be precised")
        self.Nboundaries = len(self.boundariesID)
        
    def add_inboundary(self, mesh_domain, in_domains, thickness, mat_pty=None, mat_file=None, mat_perm=None, ID=None):
        """
        add new internal boundary or update if mesh_domain already exists
        
        Parameters
        ----------
        mesh_domain     : int
            Mesh ID of the boundary (should be 2D)
        in_domains       : list
            List of mesh IDs of the domain (if 3D group) or internal boundary (if 2D group) 
            inside the boundary 
        thickness       : float
            thicness of the internal boundary
        mat_pty         : str, float or list[3]
            Material property, either mat_file or mat_perm
        mat_file        : str
            Material filename (see fmod.material.py) if None use mat_perm,
                by default None
        mat_perm        : float or list[3]
            Material permitivity, if float isotropic mat, if list[3] anisotropic mat
                if None set to 1 (and mat_file is None), by default None
        """
        if mat_pty is not None:
            mat_pty = mat_pty
        elif mat_file is not None:
            mat_pty = mat_file
        elif mat_perm is not None:
            mat_pty = mat_perm
        else:
            mat_pty = 1
            rise_warning('neither mat_file or mat_perm set, set to default permitivity 1 S/m')

        self.inbound = True
        if not mesh_domain in self.inboundaries_list:
            IDibound= self.__update_ID_list('inbound', mesh_domain)
            self.inboundaries_list[IDibound] = {'mesh_domain': mesh_domain, 'in_domains': in_domains,\
            'thickness': thickness, 'mat_pty': mat_pty}
            # Compute corresponding domain for mixed space
            for j in self.domains_list:
                domain = self.domains_list[j]
                # Domain inside the boundary
                if domain['mesh_domain'] in in_domains:
                    domain['mixed_domain'] += [domain['mesh_domain']]
                    if domain['mesh_domain'] == domain['mixed_domain'][0]:
                        domain['mixed_domain'][0] = mesh_domain
                # Domain inside the boundary but surounded by another internal boundary
                elif domain['mesh_domain']+1 in in_domains:
                    domain['mixed_domain'] += [domain['mesh_domain']+1]
                    if domain['mesh_domain'] == domain['mixed_domain'][0]:
                        domain['mixed_domain'][0] = mesh_domain
                # Domain outside the boundary
                else:
                    domain['mixed_domain'] += [mesh_domain]
        else:
            self.inboundaries_list[mesh_domain]['thickness'] = thickness
            self.inboundaries_list[mesh_domain]['mat_pty'] = mat_pty
        self.mat_pty_map[mesh_domain] = mat_pty
        self.Ninboundaries = len(self.inboundariesID)
    
    def get_mixedspace_domain(self, i_space=None, i_domain=None):
        """
        
        """
        if not self.inbound:
                if i_domain is None:
                    return self.domains_list
                else:
                    return i_domain
        else:
            if i_space is None:
                if i_domain is None:
                    md = [self.domains_list[k]['mixed_domain'] for k in self.domains_list]
                else:
                    md = self.domains_list[i_domain]['mixed_domain']
            else:
                if i_domain is None:
                    md = [self.domains_list[k]['mixed_domain'][i_space] for k in self.domains_list]
                else:
                    md = self.domains_list[i_domain]['mixed_domain'][i_space]
            return md
    
    def get_mixedspace_mat_pty(self, i_space=None, i_domain=None):
        """
        
        """
        if self.inbound:
            md = self.get_mixedspace_domain(i_space, i_domain)
            S = np.shape(md)
            if i_space is None:
                if i_domain is None:
                    mmf = [[self.mat_pty_map[md[i][j]] for i in range(S[0])] for j in range(S[1])]
                else:
                    mmf = [self.mat_pty_map[md[i]] for i in range(S[0])]
            else:
                if i_domain is None:
                    mmf = [self.mat_pty_map[md[i]] for i in range(S[0])]
                else:
                    mmf = self.mat_pty_map[md]
            return mmf

    def get_space_of_domain(self, i_domain):
        if i_domain==0:
            return 0
        for i in range(len(self.domains_list[i_domain]['mixed_domain'])):
            if self.domains_list[i_domain]['mixed_domain'][i] == i_domain:
                return(i)
        rise_error("Domain "  + str(i_domain) + " is in no space")

    def get_spaces_of_ibound(self, i_ibound):
        """
        retrun the surounding spaces of an internal boundary.
        !! Caution work for NerveMshCreator only : see if it can be generalized
        """
        in_space = self.get_space_of_domain(i_ibound-1)
        out_space = 0
        if i_ibound > 100:  # Should be a perineurium
            out_space = 0    
        else:               # Should be an axon membrane
            for i in self.inboundaries_list:
                if i_ibound in self.inboundaries_list[i]['in_domains']:
                    out_space = self.get_space_of_domain(i-1)
        return in_space, out_space
    

    def print_mixedspace_domain(self):
        doms = self.get_mixedspace_domain()
        print('spaces:   ',[k for k in range(len(doms[0]))])
        for i, dom in enumerate(doms):
            print("domain "+str(i)+": ", dom)

    def print_mixedspace_mat_pty(self):        
        mats = self.get_mixedspace_mat_pty()
        print('domains:   ',[k for k in range(len(mats[0]))])
        for i, mat in enumerate(mats):
            print("space "+str(i)+": ",mat)



if PostProc_Filtering is not None:
    filter_freq(sim_results, 'V_mem', PostProc_Filtering)
rasterize(sim_results, 'V_mem')
if not(save_V_mem) and record_V_mem:
    remove_key(sim_results, 'V_mem', verbose=verbose)



from ..backend.log_interface import rise_error, rise_warning, pass_info
from ..backend.file_handler import json_dump
from ..backend.NRV_Class import NRV_class
from .CostFunctions import CostFunction

class Problem(NRV_class):
    '''
    Problem Class

    A class to describe problems that should be optimized with the NRV Framework.
    The problem should be described with a simulation and a cost, using the object
    CostFunction, and various optimization algorithms can be used to find optimal solution.

    This class is abstract and is not supposed to be used directly by the end user. NRV can
    handle two types of problems:
        - problems where a geometric parameter can be optimized: please refer to ... 
        - problems where the waveform can be optimized: please refer to ...
    '''
    def __init__(self):
        self._CostFunction = None
        self._Optimizer = None

    # Handling the CostFunction attribute
    @property
    def CostFunction(self):
        '''
        Cost function of a Problem,
        the cost function should be a CosFunction object, it should return a scalar.
        NRV function should be prefered'''
        return self._CostFunction
    
    @CostFunction.setter
    def CostFunction(self, cost_function):
        # need to add a verification that the cost function is a scallar and so on
        self._CostFunction = cost_function

    @CostFunction.deleter
    def CostFunction(self):
        self._CostFunction = None

    def compute_cost(self, X):
        return self._CostFunction(X)

    @property
    def Optimizer(self):
        '''
        Optimizer of the problem,
        the Optimizer should be an Optimizer object. It has reference to optimization
        methods and constraints
        '''
        return self._Optimizer

    @Optimizer.setter
    def Optimizer(self, optimizer):
        self._Optimizer = optimizer

    @Optimizer.deleter
    def Optmizer(self):
        self._Optimizer = None

    # Call method is where the magic happens
    def __call__(self):
        pass

    #additional methods
    def context_and_cost(self, context_func, cost_func):
        self.CostFunction = CostFunction(context_func, cost_func)

    def autoset_optimizer(self):
        pass

    
    



nrv.rasterize(sim_results,'V_mem')
nrv.remove_key(sim_results,'V_mem', verbose=verbose)

plt.figure()
plt.scatter(sim_results['V_mem_raster_time'],sim_results['V_mem_raster_x_position'])
plt.xlabel('time (ms)')
plt.ylabel('position along the axon($\mu m$)')
title = 'Axon '+str(k)+', myelination is '+str(sim_results['myelinated'])+', '+str(sim_results['diameter'])+' um diameter'
plt.title(title)
plt.xlim(0,sim_results['tstop'])
plt.ylim(0,sim_results['L'])
plt.tight_layout()
fig_name = folder_name + '/Activity_axon_'+str(k)+'.pdf'
plt.savefig(fig_name)
plt.close()


from setuptools import setup, find_packages
from unittest.mock import MagicMock    # mock imports
import sys

# prevent from unnistalled requierements for nrv
deps = ('mph', 'neuron', 'icecream', 'numba', 'mpi4py', 'scipy', 
        'numpy', 'ezdxf', 'dolfinx', 'petsc4py', 'ufl', 'gmsh', 'dolfinx.io', 
        'petsc4py.PETSc', 'dolfinx.fem', 'dolfinx.fem.petsc', 'dolfinx.io.utils',
        'scipy.interpolate', 'scipy.special', 'dolfinx.io.gmshio', 'dolfinx.geometry',
        'dolfinx.mesh', 'numpy.linalg',  'matplotlib', 'matplotlib.pyplot', 'numpy.core', 
        'numpy.core._multiarray_umath', 'matplotlib._path', 'scipy.stats', 'scipy.optimize',
        'matplotlib.pylab', 'pylab', 'scipy.spatial', 'scipy.sparse', 'scipy.sparse.csgraph')
for package in deps:
    sys.modules[package] = MagicMock()

import nrv

setup(
   name='nrv-py',
   version=nrv.__version__,
   description=nrv.__project__,
   long_description = 'file: README.md',
   author=nrv.__contributors__,
   packages=['nrv',
             'nrv._misc',
             'nrv._misc.OTF_PP',
             'nrv._misc.comsol_templates',
             'nrv._misc.geom',
             'nrv._misc.log',
             'nrv._misc.materials',
             'nrv._misc.mods',
             'nrv._misc.pops',
             'nrv._misc.ppops',
             'nrv._misc.stats',
             'nrv.backend',
             'nrv.fmod',
             'nrv.fmod.FEM',
             'nrv.fmod.FEM.fenics_utils',
             'nrv.fmod.FEM.mesh_creator',
             'nrv.nmod',
             'nrv.optim',
             'nrv.utils',
             'nrv.utils.cell',
             'nrv.utils.fascicle'
             ],
   package_data={'nrv':['nrv2calm'],
                 'nrv._misc':['NRV.ini'],
                 'nrv._misc.comsol_templates':['*.mph'],
                 'nrv._misc.geom':['*.dxf', '*.png'],
                 'nrv._misc.log':['NRV.log'],
                 'nrv._misc.materials':['*.mat'],
                 'nrv._misc.mods':['*.mod'],
                 'nrv._misc.pops':['*.pop'],
                 'nrv._misc.ppops':['*.ppop'],
                 'nrv._misc.stats':['*.csv']},
   include_package_data = True,
   url = 'https://github.com/fkolbl/NRV',
   classifiers =[
    'Programming Language :: Python :: 3',
    'Operating System :: OS Independent'
     ],
    install_requires=['mph', 'numpy','scipy','neuron','matplotlib','numba','ezdxf','icecream'], #external packages as dependencies
    python_requires = '>=3.6'
)



"""
Access and modify NRV Parameters
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""

from abc import ABCMeta, abstractmethod
import numpy as np
import sys
from copy import deepcopy
from numpy import iterable
from .file_handler import json_dump, json_load
from .parameters import parameters
from .log_interface import rise_error, rise_warning, pass_info

def is_NRV_class(x):
    return isinstance(x, NRV_class)

def is_NRV_class_list(x):
    if iterable(x):
        for xi in x:
            if not is_NRV_class(xi):
                return False
        return True
    return False

def is_NRV_dict(x):
    if isinstance(x, dict):
        if 'nrv_type' in x:
            return True
    return False 

def is_NRV_dict_list(x):
    if iterable(x):
        for xi in x:
            if not (is_NRV_dict(xi)):
                return False
        return True
    return False




class NRV_class(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self):
        """
        """
        self.__NRVObject__ = True
        self.nrv_type = self.__class__.__name__
        if parameters.get_nrv_verbosity()>=4:
            pass_info(self.nrv_type, ' initialized')

    def __del__(self):
        if parameters.get_nrv_verbosity()>=4:
            pass_info(self.nrv_type, ' deleted')        

    def save(self,save=False, fname='nrv_save.json', blacklist={}):
        bl = {}
        for key in blacklist:
            if key in self.__dict__:
                bl[key] = self.__dict__.pop(key)
        key_dic = deepcopy(self.__dict__)
        for key in bl:
            self.__dict__[key] = bl[key]
        for key in key_dic:
            if is_NRV_class(key_dic[key]):
                key_dic[key] = key_dic[key].save()
            elif is_NRV_class_list(key_dic[key]):
                for i in range(len(key_dic[key])):
                    key_dic[key][i] = key_dic[key][i].save()

        if save:
            json_dump(key_dic, fname)
        return key_dic
        
    def load(self, data, blacklist={}):
        if type(data) == str:
            key_dic = json_load(data)
        else: 
            key_dic = data
        for key in self.__dict__:
            if key in key_dic and key not in blacklist:
                if is_NRV_dict(key_dic[key]) or is_NRV_dict_list(key_dic[key]):
                    self.__dict__[key] = load_any(key_dic[key])
                elif isinstance(self.__dict__[key], np.ndarray):
                    self.__dict__[key] = np.array(key_dic[key])
                else:
                    self.__dict__[key] = key_dic[key]


def load_any(data, **kwargs):
    if type(data) == str:
        key_dic = json_load(data)
    else: 
        key_dic = data
    
    if is_NRV_class(key_dic) or is_NRV_class_list(key_dic):
        nrv_obj = key_dic
    
    elif is_NRV_dict(key_dic):
        nrv_type = key_dic['nrv_type']
        nrv_obj = eval("sys.modules['nrv']."+nrv_type)()
        nrv_obj.load(key_dic,**kwargs)
    elif is_NRV_dict_list(key_dic):
        nrv_obj = []
        for i in key_dic:
            nrv_obj += [load_any(i, **kwargs)]
    return nrv_obj
